==> flutter_webrtc-0.9.1.tar.gz: common/darwin/Classes/FlutterRTCDesktopCapturer.m <==
#import <objc/runtime.h>

#import "FlutterRTCDesktopCapturer.h"

#if TARGET_OS_IPHONE
#import <ReplayKit/ReplayKit.h>
#import "FlutterRPScreenRecorder.h"
#import "FlutterBroadcastScreenCapturer.h"
#endif

#if TARGET_OS_OSX
RTCDesktopMediaList *_screen = nil;
RTCDesktopMediaList *_window = nil;
NSArray<RTCDesktopSource *>* _captureSources;
FlutterEventSink _eventSink = nil;
FlutterEventChannel* _eventChannel = nil;
#endif

@implementation FlutterWebRTCPlugin (DesktopCapturer)

-(void)getDisplayMedia:(NSDictionary *)constraints
                result:(FlutterResult)result {
    NSString *mediaStreamId = [[NSUUID UUID] UUIDString];
    RTCMediaStream *mediaStream = [self.peerConnectionFactory mediaStreamWithStreamId:mediaStreamId];
    RTCVideoSource *videoSource = [self.peerConnectionFactory videoSourceForScreenCast:YES];
    NSString *trackUUID = [[NSUUID UUID] UUIDString];

#if TARGET_OS_IPHONE
 BOOL useBroadcastExtension = false;
    id videoConstraints = constraints[@"video"];
    if ([videoConstraints isKindOfClass:[NSDictionary class]]) {
       // constraints.video.deviceId
        useBroadcastExtension = [((NSDictionary *)videoConstraints)[@"deviceId"] isEqualToString:@"broadcast"];
    }
    
    id screenCapturer;
    
    if(useBroadcastExtension){
        screenCapturer = [[FlutterBroadcastScreenCapturer alloc] initWithDelegate:videoSource];
    } else {
        screenCapturer = [[FlutterRPScreenRecorder alloc] initWithDelegate:videoSource];
    }
    
    [screenCapturer startCapture];
    NSLog(@"start %@ capture", useBroadcastExtension ? @"broadcast" : @"replykit");
        
    self.videoCapturerStopHandlers[trackUUID] = ^(CompletionHandler handler) {
        NSLog(@"stop %@ capture, trackID %@", useBroadcastExtension ? @"broadcast" : @"replykit", trackUUID);
        [screenCapturer stopCaptureWithCompletionHandler:handler];
    };

    if(useBroadcastExtension) {
        NSString *extension = [[[NSBundle mainBundle] infoDictionary] valueForKey: kRTCScreenSharingExtension];
        if(extension) {
            RPSystemBroadcastPickerView *picker = [[RPSystemBroadcastPickerView alloc] init];
            picker.preferredExtension = extension;
            picker.showsMicrophoneButton = false;
            
            SEL selector = NSSelectorFromString(@"buttonPressed:");
            if([picker respondsToSelector:selector]) {
                [picker performSelector:selector withObject:nil];
            }
        }
    }
#endif
    
#if TARGET_OS_OSX
/* example for constraints:
    {
        'audio': false,
        'video": {
            'deviceId':  {'exact': sourceId},
            'mandatory': {
                'frameRate': 30.0
            },
        }
    }
*/
    NSString *sourceId = nil;
    BOOL useDefaultScreen = NO;
    NSInteger fps = 30;
    id videoConstraints = constraints[@"video"];
    if([videoConstraints isKindOfClass:[NSNumber class]] && [videoConstraints boolValue] == YES) {
        useDefaultScreen = YES;
    } else if ([videoConstraints isKindOfClass:[NSDictionary class]]) {
        NSDictionary *deviceId = videoConstraints[@"deviceId"];
        if (deviceId != nil && [deviceId isKindOfClass:[NSDictionary class]]) {
            if(deviceId[@"exact"] != nil) {
                sourceId = deviceId[@"exact"];
                if(sourceId == nil) {
                    result(@{@"error": @"No deviceId.exact found"});
                    return;
                }
            }
        } else {
            // fall back to default screen if no deviceId is specified
            useDefaultScreen = YES;
        }
        id mandatory = videoConstraints[@"mandatory"];
        if (mandatory != nil && [mandatory isKindOfClass:[NSDictionary class]]) {
            id frameRate = mandatory[@"frameRate"];
            if (frameRate != nil && [frameRate isKindOfClass:[NSNumber class]]) {
                fps = [frameRate integerValue];
            }
        }
    }
    RTCDesktopCapturer *desktopCapturer;
    RTCDesktopSource *source = nil;
    if(useDefaultScreen){
        desktopCapturer  = [[RTCDesktopCapturer alloc] initWithDefaultScreen:self captureDelegate:videoSource];
    } else {
         source = [self getSourceById:sourceId];
        if(source == nil) {
            result(@{@"error":  [NSString stringWithFormat:@"No source found for id: %@",sourceId]});
            return;
        }
        desktopCapturer  = [[RTCDesktopCapturer alloc] initWithSource:source delegate:self captureDelegate:videoSource];
    }
    [desktopCapturer startCaptureWithFPS:fps];
    NSLog(@"start desktop capture: sourceId: %@, type: %@, fps: %lu", sourceId, source.sourceType == RTCDesktopSourceTypeScreen ? @"screen" : @"window", fps);

    self.videoCapturerStopHandlers[trackUUID] = ^(CompletionHandler handler) {
        NSLog(@"stop desktop capture: sourceId: %@, type: %@, trackID %@", sourceId, source.sourceType == RTCDesktopSourceTypeScreen ? @"screen" : @"window", trackUUID);
        [desktopCapturer stopCapture];
        handler();
    };
#endif

    RTCVideoTrack *videoTrack = [self.peerConnectionFactory videoTrackWithSource:videoSource trackId:trackUUID];
    [mediaStream addVideoTrack:videoTrack];

    NSMutableArray *audioTracks = [NSMutableArray array];
    NSMutableArray *videoTracks = [NSMutableArray array];

    for (RTCVideoTrack *track in mediaStream.videoTracks) {
        [self.localTracks setObject:track forKey:track.trackId];
        [videoTracks addObject:@{@"id": track.trackId, @"kind": track.kind, @"label": track.trackId, @"enabled": @(track.isEnabled), @"remote": @(YES), @"readyState": @"live"}];
    }

    self.localStreams[mediaStreamId] = mediaStream;
    result(@{@"streamId": mediaStreamId, @"audioTracks" : audioTracks, @"videoTracks" : videoTracks });
}

-(void)getDesktopSources:(NSDictionary *)argsMap
             result:(FlutterResult)result {
#if TARGET_OS_OSX
    NSLog(@"getDesktopSources");

    NSArray *types = [argsMap objectForKey:@"types"];
    if (types == nil) {
        result([FlutterError errorWithCode:@"ERROR"
                                   message:@"types is required"
                                   details:nil]);
        return;
    }

    if(![self buildDesktopSourcesListWithTypes:types forceReload:YES result:result]) {
        NSLog(@"getDesktopSources failed.");
        return;
    }

    NSMutableArray *sources = [NSMutableArray array];
    NSEnumerator *enumerator = [_captureSources objectEnumerator];
    RTCDesktopSource *object;
    while ((object = enumerator.nextObject) != nil) {
        NSData *data = nil;
        if([object thumbnail]) {
            data = [[NSData alloc] init];
            NSImage *resizedImg = [self resizeImage:[object thumbnail] forSize:NSMakeSize(140, 140)];
            data = [resizedImg TIFFRepresentation];
        }
        [sources addObject:@{
                             @"id": object.sourceId,
                             @"name": object.name,
                             @"thumbnailSize": @{@"width": @0, @"height": @0},
                             @"type": object.sourceType == RTCDesktopSourceTypeScreen? @"screen" : @"window",
                             @"thumbnail": data,
                             }];
    }
    result(@{@"sources": sources});
#else
    result([FlutterError errorWithCode:@"ERROR"
                               message:@"Not supported on iOS"
                               details:nil]);
#endif
}

-(void)getDesktopSourceThumbnail:(NSDictionary *)argsMap
             result:(FlutterResult)result {
#if TARGET_OS_OSX
    NSLog(@"getDesktopSourceThumbnail");
    NSString* sourceId = argsMap[@"sourceId"];
    RTCDesktopSource *object = [self getSourceById:sourceId];
    if(object == nil) {
        result(@{@"error": @"No source found"});
        return;
    }
    NSImage *image = [object UpdateThumbnail];
    if(image != nil) {
        NSImage *resizedImg = [self resizeImage:image forSize:NSMakeSize(140, 140)];
        NSData *data = [resizedImg TIFFRepresentation];
        result(data);
    } else {
        result(@{@"error": @"No thumbnail found"});
    }
    
#else
    result([FlutterError errorWithCode:@"ERROR"
                               message:@"Not supported on iOS"
                               details:nil]);
#endif
}

-(void)updateDesktopSources:(NSDictionary *)argsMap
             result:(FlutterResult)result {
#if TARGET_OS_OSX
    NSLog(@"updateDesktopSources");
    NSArray *types = [argsMap objectForKey:@"types"];
    if (types == nil) {
        result([FlutterError errorWithCode:@"ERROR"
                                   message:@"types is required"
                                   details:nil]);
        return;
    }
    if(![self buildDesktopSourcesListWithTypes:types forceReload:NO result:result]) {
        NSLog(@"updateDesktopSources failed.");
        return;
    }
    result(@{@"result": @YES});
#else
    result([FlutterError errorWithCode:@"ERROR"
                               message:@"Not supported on iOS"
                               details:nil]);
#endif
}

#if TARGET_OS_OSX
- (NSImage*)resizeImage:(NSImage*)sourceImage forSize:(CGSize)targetSize {
    CGSize imageSize = sourceImage.size;
    CGFloat width = imageSize.width;
    CGFloat height = imageSize.height;
    CGFloat targetWidth = targetSize.width;
    CGFloat targetHeight = targetSize.height;
    CGFloat scaleFactor = 0.0;
    CGFloat scaledWidth = targetWidth;
    CGFloat scaledHeight = targetHeight;
    CGPoint thumbnailPoint = CGPointMake(0.0,0.0);

    if (CGSizeEqualToSize(imageSize, targetSize) == NO) {
        CGFloat widthFactor = targetWidth / width;
        CGFloat heightFactor = targetHeight / height;

        // scale to fit the longer
        scaleFactor = (widthFactor>heightFactor)?widthFactor:heightFactor;
        scaledWidth  = ceil(width * scaleFactor);
        scaledHeight = ceil(height * scaleFactor);

        // center the image
        if (widthFactor > heightFactor) {
            thumbnailPoint.y = (targetHeight - scaledHeight) * 0.5;
        } else if (widthFactor < heightFactor) {
            thumbnailPoint.x = (targetWidth - scaledWidth) * 0.5;
        }
    }

    NSImage *newImage = [[NSImage alloc] initWithSize:NSMakeSize(scaledWidth, scaledHeight)];
    CGRect thumbnailRect = {thumbnailPoint, {scaledWidth, scaledHeight}};
    NSRect imageRect = NSMakeRect(0.0, 0.0, width, height);

    [newImage lockFocus];
    [sourceImage drawInRect:thumbnailRect fromRect:imageRect operation:NSCompositeCopy fraction:1.0];
    [newImage unlockFocus];

    return newImage;
}

-(RTCDesktopSource *)getSourceById:(NSString *)sourceId {
    NSEnumerator *enumerator = [_captureSources objectEnumerator];
    RTCDesktopSource *object;
    while ((object = enumerator.nextObject) != nil) {
        if([sourceId isEqualToString:object.sourceId]) {
            return object;
        }
    }
    return nil;
}

- (BOOL)buildDesktopSourcesListWithTypes:(NSArray *)types forceReload:(BOOL)forceReload result:(FlutterResult)result {
    BOOL captureWindow = NO;
    BOOL captureScreen = NO;
    _captureSources = [NSMutableArray array];

    NSEnumerator *typesEnumerator = [types objectEnumerator];
    NSString *type;
    while ((type = typesEnumerator.nextObject) != nil) {
        if ([type isEqualToString:@"screen"]) {
            captureScreen = YES;
        } else if ([type isEqualToString:@"window"]) {
            captureWindow = YES;
        } else {
            result([FlutterError errorWithCode:@"ERROR"
                                       message:@"Invalid type"
                                       details:nil]);
            return NO;
        }
    }

    if(!captureWindow && !captureScreen) {
        result([FlutterError errorWithCode:@"ERROR"
                                   message:@"At least one type is required"
                                   details:nil]);
        return NO;
    }

     if(captureWindow) {
        if(!_window) _window = [[RTCDesktopMediaList alloc] initWithType:RTCDesktopSourceTypeWindow delegate:self];
         [_window UpdateSourceList:forceReload updateAllThumbnails:YES];
        NSArray<RTCDesktopSource *>* sources = [_window getSources];
        _captureSources = [_captureSources arrayByAddingObjectsFromArray:sources];
    }
    if(captureScreen) {
        if(!_screen) _screen = [[RTCDesktopMediaList alloc] initWithType:RTCDesktopSourceTypeScreen  delegate:self];
        [_screen UpdateSourceList:forceReload updateAllThumbnails:YES];
        NSArray<RTCDesktopSource *>* sources = [_screen getSources];
        _captureSources = [_captureSources arrayByAddingObjectsFromArray:sources];
    }
    NSLog(@"captureSources: %lu", [_captureSources count]);
    return YES;
}

-(void) enableDesktopCapturerEventChannel:(nonnull NSObject<FlutterBinaryMessenger> *)messenger {
    if(_eventChannel == nil) {
        _eventChannel = [FlutterEventChannel
                                            eventChannelWithName:@"FlutterWebRTC/desktopSourcesEvent"
                                            binaryMessenger:messenger];
        [_eventChannel setStreamHandler:self];
    }
}

#pragma mark - FlutterStreamHandler methods

#pragma clang diagnostic ignored "-Wobjc-protocol-method-implementation"
- (FlutterError* _Nullable)onCancelWithArguments:(id _Nullable)arguments {
    _eventSink = nil;
    return nil;
}

#pragma clang diagnostic ignored "-Wobjc-protocol-method-implementation"
- (FlutterError* _Nullable)onListenWithArguments:(id _Nullable)arguments
                                       eventSink:(nonnull FlutterEventSink)sink {
    _eventSink = sink;
    return nil;
}

#pragma mark - RTCDesktopMediaListDelegate delegate

#pragma clang diagnostic ignored "-Wobjc-protocol-method-implementation"
- (void)didDesktopSourceAdded:(RTC_OBJC_TYPE(RTCDesktopSource) *)source {
    //NSLog(@"didDesktopSourceAdded: %@, id %@", source.name, source.sourceId);
    if(_eventSink) {
        NSImage *image = [source UpdateThumbnail];
        NSData *data = [[NSData alloc] init];
        if(image != nil) {
            NSImage *resizedImg = [self resizeImage:image forSize:NSMakeSize(140, 140)];
            data = [resizedImg TIFFRepresentation];
        }
        _eventSink(@{
            @"event": @"desktopSourceAdded",
            @"id": source.sourceId,
            @"name": source.name,
            @"thumbnailSize": @{@"width": @0, @"height": @0},
            @"type": source.sourceType == RTCDesktopSourceTypeScreen? @"screen" : @"window",
            @"thumbnail": data
        });
    }
}

#pragma clang diagnostic ignored "-Wobjc-protocol-method-implementation"
- (void)didDesktopSourceRemoved:(RTC_OBJC_TYPE(RTCDesktopSource) *) source {
   //NSLog(@"didDesktopSourceRemoved: %@, id %@", source.name, source.sourceId);
    if(_eventSink) {
        _eventSink(@{
            @"event": @"desktopSourceRemoved",
            @"id": source.sourceId,
        });
    }
}

#pragma clang diagnostic ignored "-Wobjc-protocol-method-implementation"
- (void)didDesktopSourceNameChanged:(RTC_OBJC_TYPE(RTCDesktopSource) *) source {
    //NSLog(@"didDesktopSourceNameChanged: %@, id %@", source.name, source.sourceId);
    if(_eventSink) {
        _eventSink(@{
            @"event": @"desktopSourceNameChanged",
            @"id": source.sourceId,
            @"name": source.name,
        });
    }
}

#pragma clang diagnostic ignored "-Wobjc-protocol-method-implementation"
- (void)didDesktopSourceThumbnailChanged:(RTC_OBJC_TYPE(RTCDesktopSource) *) source {
    //NSLog(@"didDesktopSourceThumbnailChanged: %@, id %@", source.name, source.sourceId);
    if(_eventSink) {
        NSImage *resizedImg = [self resizeImage:[source thumbnail] forSize:NSMakeSize(140, 140)];
        NSData *data = [resizedImg TIFFRepresentation];
        _eventSink(@{
            @"event": @"desktopSourceThumbnailChanged",
            @"id": source.sourceId,
            @"thumbnail": data 
        });
    }
}

#pragma mark - RTCDesktopCapturerDelegate delegate

-(void)didSourceCaptureStart:(RTCDesktopCapturer *) capturer {
    NSLog(@"didSourceCaptureStart");
}

-(void)didSourceCapturePaused:(RTCDesktopCapturer *) capturer {
    NSLog(@"didSourceCapturePaused");
}

-(void)didSourceCaptureStop:(RTCDesktopCapturer *) capturer {
    NSLog(@"didSourceCaptureStop");
}

-(void)didSourceCaptureError:(RTCDesktopCapturer *) capturer{
    NSLog(@"didSourceCaptureError");
}

#endif

@end

==> flutter_webrtc-0.9.10.tar.gz: common/darwin/Classes/FlutterRTCDesktopCapturer.m <==
#import <objc/runtime.h>

#import "FlutterRTCDesktopCapturer.h"

#if TARGET_OS_IPHONE
#import <ReplayKit/ReplayKit.h>
#import "FlutterRPScreenRecorder.h"
#import "FlutterBroadcastScreenCapturer.h"
#endif

#if TARGET_OS_OSX
RTCDesktopMediaList *_screen = nil;
RTCDesktopMediaList *_window = nil;
NSArray<RTCDesktopSource *>* _captureSources;
#endif

@implementation FlutterWebRTCPlugin (DesktopCapturer)

-(void)getDisplayMedia:(NSDictionary *)constraints
                result:(FlutterResult)result {
    NSString *mediaStreamId = [[NSUUID UUID] UUIDString];
    RTCMediaStream *mediaStream = [self.peerConnectionFactory mediaStreamWithStreamId:mediaStreamId];
    RTCVideoSource *videoSource = [self.peerConnectionFactory videoSourceForScreenCast:YES];
    NSString *trackUUID = [[NSUUID UUID] UUIDString];

#if TARGET_OS_IPHONE
 BOOL useBroadcastExtension = false;
    id videoConstraints = constraints[@"video"];
    if ([videoConstraints isKindOfClass:[NSDictionary class]]) {
       // constraints.video.deviceId
        useBroadcastExtension = [((NSDictionary *)videoConstraints)[@"deviceId"] isEqualToString:@"broadcast"];
    }
    
    id screenCapturer;
    
    if(useBroadcastExtension){
        screenCapturer = [[FlutterBroadcastScreenCapturer alloc] initWithDelegate:videoSource];
    } else {
        screenCapturer = [[FlutterRPScreenRecorder alloc] initWithDelegate:videoSource];
    }
    
    [screenCapturer startCapture];
    NSLog(@"start %@ capture", useBroadcastExtension ? @"broadcast" : @"replykit");
        
    self.videoCapturerStopHandlers[trackUUID] = ^(CompletionHandler handler) {
        NSLog(@"stop %@ capture, trackID %@", useBroadcastExtension ? @"broadcast" : @"replykit", trackUUID);
        [screenCapturer stopCaptureWithCompletionHandler:handler];
    };

    if(useBroadcastExtension) {
        NSString *extension = [[[NSBundle mainBundle] infoDictionary] valueForKey: kRTCScreenSharingExtension];
        if(extension) {
            RPSystemBroadcastPickerView *picker = [[RPSystemBroadcastPickerView alloc] init];
            picker.preferredExtension = extension;
            picker.showsMicrophoneButton = false;
            
            SEL selector = NSSelectorFromString(@"buttonPressed:");
            if([picker respondsToSelector:selector]) {
                [picker performSelector:selector withObject:nil];
            }
        }
    }
#endif
    
#if TARGET_OS_OSX
/* example for constraints:
    {
        'audio': false,
        'video": {
            'deviceId':  {'exact': sourceId},
            'mandatory': {
                'frameRate': 30.0
            },
        }
    }
*/
    NSString *sourceId = nil;
    BOOL useDefaultScreen = NO;
    NSInteger fps = 30;
    id videoConstraints = constraints[@"video"];
    if([videoConstraints isKindOfClass:[NSNumber class]] && [videoConstraints boolValue] == YES) {
        useDefaultScreen = YES;
    } else if ([videoConstraints isKindOfClass:[NSDictionary class]]) {
        NSDictionary *deviceId = videoConstraints[@"deviceId"];
        if (deviceId != nil && [deviceId isKindOfClass:[NSDictionary class]]) {
            if(deviceId[@"exact"] != nil) {
                sourceId = deviceId[@"exact"];
                if(sourceId == nil) {
                    result(@{@"error": @"No deviceId.exact found"});
                    return;
                }
            }
        } else {
            // fall back to default screen if no deviceId is specified
            useDefaultScreen = YES;
        }
        id mandatory = videoConstraints[@"mandatory"];
        if (mandatory != nil && [mandatory isKindOfClass:[NSDictionary class]]) {
            id frameRate = mandatory[@"frameRate"];
            if (frameRate != nil && [frameRate isKindOfClass:[NSNumber class]]) {
                fps = [frameRate integerValue];
            }
        }
    }
    RTCDesktopCapturer *desktopCapturer;
    RTCDesktopSource *source = nil;
    if(useDefaultScreen){
        desktopCapturer  = [[RTCDesktopCapturer alloc] initWithDefaultScreen:self captureDelegate:videoSource];
    } else {
         source = [self getSourceById:sourceId];
        if(source == nil) {
            result(@{@"error":  [NSString stringWithFormat:@"No source found for id: %@",sourceId]});
            return;
        }
        desktopCapturer  = [[RTCDesktopCapturer alloc] initWithSource:source delegate:self captureDelegate:videoSource];
    }
    [desktopCapturer startCaptureWithFPS:fps];
    NSLog(@"start desktop capture: sourceId: %@, type: %@, fps: %lu", sourceId, source.sourceType == RTCDesktopSourceTypeScreen ? @"screen" : @"window", fps);

    self.videoCapturerStopHandlers[trackUUID] = ^(CompletionHandler handler) {
        NSLog(@"stop desktop capture: sourceId: %@, type: %@, trackID %@", sourceId, source.sourceType == RTCDesktopSourceTypeScreen ? @"screen" : @"window", trackUUID);
        [desktopCapturer stopCapture];
        handler();
    };
#endif

    RTCVideoTrack *videoTrack = [self.peerConnectionFactory videoTrackWithSource:videoSource trackId:trackUUID];
    [mediaStream addVideoTrack:videoTrack];

    [self.localTracks setObject:videoTrack forKey:trackUUID];

    NSMutableArray *audioTracks = [NSMutableArray array];
    NSMutableArray *videoTracks = [NSMutableArray array];

    for (RTCVideoTrack *track in mediaStream.videoTracks) {
        [videoTracks addObject:@{@"id": track.trackId, @"kind": track.kind, @"label": track.trackId, @"enabled": @(track.isEnabled), @"remote": @(YES), @"readyState": @"live"}];
    }

    self.localStreams[mediaStreamId] = mediaStream;
    result(@{@"streamId": mediaStreamId, @"audioTracks" : audioTracks, @"videoTracks" : videoTracks });
}

-(void)getDesktopSources:(NSDictionary *)argsMap
             result:(FlutterResult)result {
#if TARGET_OS_OSX
    NSLog(@"getDesktopSources");

    NSArray *types = [argsMap objectForKey:@"types"];
    if (types == nil) {
        result([FlutterError errorWithCode:@"ERROR"
                                   message:@"types is required"
                                   details:nil]);
        return;
    }

    if(![self buildDesktopSourcesListWithTypes:types forceReload:YES result:result]) {
        NSLog(@"getDesktopSources failed.");
        return;
    }

    NSMutableArray *sources = [NSMutableArray array];
    NSEnumerator *enumerator = [_captureSources objectEnumerator];
    RTCDesktopSource *object;
    while ((object = enumerator.nextObject) != nil) {
        /*NSData *data = nil;
        if([object thumbnail]) {
            data = [[NSData alloc] init];
            NSImage *resizedImg = [self resizeImage:[object thumbnail] forSize:NSMakeSize(320, 180)];
            data = [resizedImg TIFFRepresentation];
        }*/
        [sources addObject:@{
                             @"id": object.sourceId,
                             @"name": object.name,
                             @"thumbnailSize": @{@"width": @0, @"height": @0},
                             @"type": object.sourceType == RTCDesktopSourceTypeScreen? @"screen" : @"window",
                             //@"thumbnail": data,
                             }];
    }
    result(@{@"sources": sources});
#else
    result([FlutterError errorWithCode:@"ERROR"
                               message:@"Not supported on iOS"
                               details:nil]);
#endif
}

-(void)getDesktopSourceThumbnail:(NSDictionary *)argsMap
             result:(FlutterResult)result {
#if TARGET_OS_OSX
    NSLog(@"getDesktopSourceThumbnail");
    NSString* sourceId = argsMap[@"sourceId"];
    RTCDesktopSource *object = [self getSourceById:sourceId];
    if(object == nil) {
        result(@{@"error": @"No source found"});
        return;
    }
    NSImage *image = [object UpdateThumbnail];
    if(image != nil) {
        NSImage *resizedImg = [self resizeImage:image forSize:NSMakeSize(320, 180)];
        NSData *data = [resizedImg TIFFRepresentation];
        result(data);
    } else {
        result(@{@"error": @"No thumbnail found"});
    }
    
#else
    result([FlutterError errorWithCode:@"ERROR"
                               message:@"Not supported on iOS"
                               details:nil]);
#endif
}

-(void)updateDesktopSources:(NSDictionary *)argsMap
             result:(FlutterResult)result {
#if TARGET_OS_OSX
    NSLog(@"updateDesktopSources");
    NSArray *types = [argsMap objectForKey:@"types"];
    if (types == nil) {
        result([FlutterError errorWithCode:@"ERROR"
                                   message:@"types is required"
                                   details:nil]);
        return;
    }
    if(![self buildDesktopSourcesListWithTypes:types forceReload:NO result:result]) {
        NSLog(@"updateDesktopSources failed.");
        return;
    }
    result(@{@"result": @YES});
#else
    result([FlutterError errorWithCode:@"ERROR"
                               message:@"Not supported on iOS"
                               details:nil]);
#endif
}

#if TARGET_OS_OSX
- (NSImage*)resizeImage:(NSImage*)sourceImage forSize:(CGSize)targetSize {
    CGSize imageSize = sourceImage.size;
    CGFloat width = imageSize.width;
    CGFloat height = imageSize.height;
    CGFloat targetWidth = targetSize.width;
    CGFloat targetHeight = targetSize.height;
    CGFloat scaleFactor = 0.0;
    CGFloat scaledWidth = targetWidth;
    CGFloat scaledHeight = targetHeight;
    CGPoint thumbnailPoint = CGPointMake(0.0,0.0);

    if (CGSizeEqualToSize(imageSize, targetSize) == NO) {
        CGFloat widthFactor = targetWidth / width;
        CGFloat heightFactor = targetHeight / height;

        // scale to fit the longer
        scaleFactor = (widthFactor>heightFactor)?widthFactor:heightFactor;
        scaledWidth  = ceil(width * scaleFactor);
        scaledHeight = ceil(height * scaleFactor);

        // center the image
        if (widthFactor > heightFactor) {
            thumbnailPoint.y = (targetHeight - scaledHeight) * 0.5;
        } else if (widthFactor < heightFactor) {
            thumbnailPoint.x = (targetWidth - scaledWidth) * 0.5;
        }
    }

    NSImage *newImage = [[NSImage alloc] initWithSize:NSMakeSize(scaledWidth, scaledHeight)];
    CGRect thumbnailRect = {thumbnailPoint, {scaledWidth, scaledHeight}};
    NSRect imageRect = NSMakeRect(0.0, 0.0, width, height);

    [newImage lockFocus];
    [sourceImage drawInRect:thumbnailRect fromRect:imageRect operation:NSCompositeCopy fraction:1.0];
    [newImage unlockFocus];

    return newImage;
}

-(RTCDesktopSource *)getSourceById:(NSString *)sourceId {
    NSEnumerator *enumerator = [_captureSources objectEnumerator];
    RTCDesktopSource *object;
    while ((object = enumerator.nextObject) != nil) {
        if([sourceId isEqualToString:object.sourceId]) {
            return object;
        }
    }
    return nil;
}

- (BOOL)buildDesktopSourcesListWithTypes:(NSArray *)types forceReload:(BOOL)forceReload result:(FlutterResult)result {
    BOOL captureWindow = NO;
    BOOL captureScreen = NO;
    _captureSources = [NSMutableArray array];

    NSEnumerator *typesEnumerator = [types objectEnumerator];
    NSString *type;
    while ((type = typesEnumerator.nextObject) != nil) {
        if ([type isEqualToString:@"screen"]) {
            captureScreen = YES;
        } else if ([type isEqualToString:@"window"]) {
            captureWindow = YES;
        } else {
            result([FlutterError errorWithCode:@"ERROR"
                                       message:@"Invalid type"
                                       details:nil]);
            return NO;
        }
    }

    if(!captureWindow && !captureScreen) {
        result([FlutterError errorWithCode:@"ERROR"
                                   message:@"At least one type is required"
                                   details:nil]);
        return NO;
    }

     if(captureWindow) {
        if(!_window) _window = [[RTCDesktopMediaList alloc] initWithType:RTCDesktopSourceTypeWindow delegate:self];
         [_window UpdateSourceList:forceReload updateAllThumbnails:YES];
        NSArray<RTCDesktopSource *>* sources = [_window getSources];
        _captureSources = [_captureSources arrayByAddingObjectsFromArray:sources];
    }
    if(captureScreen) {
        if(!_screen) _screen = [[RTCDesktopMediaList alloc] initWithType:RTCDesktopSourceTypeScreen  delegate:self];
        [_screen UpdateSourceList:forceReload updateAllThumbnails:YES];
        NSArray<RTCDesktopSource *>* sources = [_screen getSources];
        _captureSources = [_captureSources arrayByAddingObjectsFromArray:sources];
    }
    NSLog(@"captureSources: %lu", [_captureSources count]);
    return YES;
}

#pragma mark - RTCDesktopMediaListDelegate delegate

#pragma clang diagnostic ignored "-Wobjc-protocol-method-implementation"
- (void)didDesktopSourceAdded:(RTC_OBJC_TYPE(RTCDesktopSource) *)source {
    //NSLog(@"didDesktopSourceAdded: %@, id %@", source.name, source.sourceId);
    if(self.eventSink) {
        NSImage *image = [source UpdateThumbnail];
        NSData *data = [[NSData alloc] init];
        if(image != nil) {
            NSImage *resizedImg = [self resizeImage:image forSize:NSMakeSize(320, 180)];
            data = [resizedImg TIFFRepresentation];
        }
        self.eventSink(@{
            @"event": @"desktopSourceAdded",
            @"id": source.sourceId,
            @"name": source.name,
            @"thumbnailSize": @{@"width": @0, @"height": @0},
            @"type": source.sourceType == RTCDesktopSourceTypeScreen? @"screen" : @"window",
            @"thumbnail": data
        });
    }
}

#pragma clang diagnostic ignored "-Wobjc-protocol-method-implementation"
- (void)didDesktopSourceRemoved:(RTC_OBJC_TYPE(RTCDesktopSource) *) source {
   //NSLog(@"didDesktopSourceRemoved: %@, id %@", source.name, source.sourceId);
    if(self.eventSink) {
        self.eventSink(@{
            @"event": @"desktopSourceRemoved",
            @"id": source.sourceId,
        });
    }
}

#pragma clang diagnostic ignored "-Wobjc-protocol-method-implementation"
- (void)didDesktopSourceNameChanged:(RTC_OBJC_TYPE(RTCDesktopSource) *) source {
    //NSLog(@"didDesktopSourceNameChanged: %@, id %@", source.name, source.sourceId);
    if(self.eventSink) {
        self.eventSink(@{
            @"event": @"desktopSourceNameChanged",
            @"id": source.sourceId,
            @"name": source.name,
        });
    }
}

#pragma clang diagnostic ignored "-Wobjc-protocol-method-implementation"
- (void)didDesktopSourceThumbnailChanged:(RTC_OBJC_TYPE(RTCDesktopSource) *) source {
    //NSLog(@"didDesktopSourceThumbnailChanged: %@, id %@", source.name, source.sourceId);
    if(self.eventSink) {
        NSImage *resizedImg = [self resizeImage:[source thumbnail] forSize:NSMakeSize(320, 180)];
        NSData *data = [resizedImg TIFFRepresentation];
        self.eventSink(@{
            @"event": @"desktopSourceThumbnailChanged",
            @"id": source.sourceId,
            @"thumbnail": data 
        });
    }
}

#pragma mark - RTCDesktopCapturerDelegate delegate

-(void)didSourceCaptureStart:(RTCDesktopCapturer *) capturer {
    NSLog(@"didSourceCaptureStart");
}

-(void)didSourceCapturePaused:(RTCDesktopCapturer *) capturer {
    NSLog(@"didSourceCapturePaused");
}

-(void)didSourceCaptureStop:(RTCDesktopCapturer *) capturer {
    NSLog(@"didSourceCaptureStop");
}

-(void)didSourceCaptureError:(RTCDesktopCapturer *) capturer{
    NSLog(@"didSourceCaptureError");
}

#endif

@end
